"""Time utils."""
import time
import datetime

DEFAULT_TIME_FORMAT = '%Y-%m-%d %H:%M:%S'


class SECONDS_IN:
    """Units of time."""
    MINUTE = 60
    HOUR = 3600
    DAY = 86400
    WEEK = 7 * 86400
    FORTNIGHT = 14 * 86400
    MONTH = 365.25 * 4 * 86400 / 12
    QTR = 365.25 * 86400 / 4
    YEAR = 365.25 * 86400


def get_timezone():
    """Get timezone."""
    return str(datetime.datetime.now().astimezone().tzinfo)


def get_unixtime():
    """Get current unixtime."""
    return (int)(time.time())


def parse_time(
    time_str,
    time_format=DEFAULT_TIME_FORMAT,
):
    """Parse time string, and return unixtime."""
    _datetime = datetime.datetime.strptime(time_str, time_format)
    return (int)(time.mktime(_datetime.timetuple()))


def format_time(
    unixtime,
    time_format=DEFAULT_TIME_FORMAT,
):
    """Format unixtime as time string."""
    _datetime = datetime.datetime.fromtimestamp(unixtime)
    return _datetime.strftime(time_format)


class StopWatch:
    """Implements StopWatch.

    .. code-block:: python

        >>> import time
        >>> from utils import timex
        >>> sw = timex.StopWatch()
        >>> time.sleep(3)
        >>> print(sw.stop('Test'))
        3930.06706237793

    """

    def __init__(self):
        """__init__."""
        self.reset()

    def reset(self):
        """Reset StopWatch."""
        self.t_start = time.time()

    def stop(self, label=""):
        """Stop StopWatch, print and return elapsed time.

        Args:
            label(str, optional): Optional label to prefix the output

        Return:
            Elapsed time in miliseconds (float)

        """
        delta_t = (time.time() - self.t_start) * 1000
        print('{label} {delta_t}ms'.format(label=label, delta_t=delta_t))
        return delta_t



"""JSON utils.

.. code-block:: python

    >>> from utils import jsonx
    >>> data = {'name': 'Alice', 'age': 20}
    >>> file_name = '/tmp/data.json'
    >>> jsonx.write(file_name, data)
    >>> data2 = jsonx.read(file_name)
    >>> data == data2
    True

"""
import json

from utils import filex


def read(file_name):
    """Read JSON from file.

    Args:
        file_name (str): file name

    Returns:
        Parsed JSON data

    """
    return json.loads(filex.read(file_name))


def write(file_name, data):
    """Write data as JSON to file.

    Args:
        file_name (str): file name
        data: data as serializable object

    """
    filex.write(file_name, json.dumps(data, indent=2))



"""Utils for reading remote files."""
import json
import logging
import ssl
import time

import requests

from bs4 import BeautifulSoup
from selenium import webdriver
from selenium.webdriver.firefox.options import Options

from utils import filex, tsv

USER_AGENT = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.14; rv:65.0)' \
    + 'Gecko/20100101 Firefox/65.0'
ENCODING = 'utf-8'
SELENIUM_SCROLL_SCRIPT = "window.scrollTo(0, document.body.scrollHeight);"
SELENIUM_SCROLL_REPEATS = 3
SELENIUM_SCROLL_WAIT_TIME = 0.5

# pylint: disable=W0212
ssl._create_default_https_context = ssl._create_unverified_context


def _read_helper(url):
    try:
        logging.debug('utils.www._read_helper: %s', url)
        resp = requests.get(url, headers={'user-agent': USER_AGENT})

        if resp.status_code != 200:
            return None

        return resp.content

    except requests.exceptions.ConnectionError:
        return None


def _read_helper_selenium(url):
    options = Options()
    options.headless = True

    browser = webdriver.Firefox(options=options)
    browser.get(url)

    for _ in range(0, SELENIUM_SCROLL_REPEATS):
        browser.execute_script(SELENIUM_SCROLL_SCRIPT)
        time.sleep(SELENIUM_SCROLL_WAIT_TIME)

    content = browser.page_source
    browser.quit()
    return content


def read(url, use_selenium=False):
    """Read url.

    Args:
        url (str): URL

    Return:
        Contents at URL

    """
    if use_selenium:
        content = _read_helper_selenium(url)
    else:
        content = _read_helper(url)
        if content:
            content = content.decode(ENCODING)

    return content


def read_json(url):
    """Read JSON content from url.

    Args:
        url (str): URL

    Return:
        JSON parsed data at URL

    """
    try:
        return json.loads(read(url))
    except TypeError:
        return None


def read_tsv(url):
    """Read TSV content from url.

    Args:
        url (str): URL

    Return:
        TSV parsed data at URL
    """
    csv_lines = read(url).split('\n')
    return tsv._read_helper(csv_lines)


def download_binary(url, file_name):
    """Download binary.

    Args:
        url (str): URL
        file_name (str): file name for output
    """
    content = _read_helper(url)
    filex.write(file_name, content, 'wb')
    logging.debug(
        'Wrote %dB from %s to %s',
        len(content),
        url,
        file_name,
    )


def exists(url):
    """Check if URL exists."""
    try:
        response = requests.head(url, timeout=1)
    except requests.exceptions.ConnectTimeout:
        return False
    return response.status_code == requests.codes.ok


def get_all_urls(root_url):
    """Get all URLs linked to a webpage."""
    soup = BeautifulSoup(read(root_url), 'html.parser')
    urls = []
    for a_link in soup.find_all('a', href=True):
        urls.append(a_link['href'])
    logging.debug('Found %d links on %s', len(urls), root_url)
    return urls



"""Utils for caching functions."""
import hashlib
import json
import os
import threading
import time
from functools import wraps

from utils import filex

CACHE_DIR = '/tmp/cache'
CACHE_DEFAULT_TIMEOUT = 60


class _Cache:
    """Implements base cache logic.

    Data is cached both in memory and, for persistance, on disk. On *get*
    requests, memory is first inspected. If cache key is not found there,
    disk is inspected.

    Args:
        cache_name(str): Name to identify cache.
        timeout (int): timeout in seconds

    Note:
        Used by decorator *cache* (See below).

    """

    @staticmethod
    def __get_hash(cache_key):
        return hashlib.md5(cache_key.encode()).hexdigest()

    __dir = '/tmp/cache'
    __store = {}
    __lock_map_lock = threading.Lock()
    __lock_map = {}
    __cache_name = 'new_cache'

    def __init__(self, cache_name, timeout=CACHE_DEFAULT_TIMEOUT):
        """Implement class constructor."""
        self.__cache_name = cache_name
        self.__timeout = timeout
        os.system('mkdir -p %s' % self.__get_dir())

    def __get_dir(self):
        return '%s/%s' % (CACHE_DIR, self.__cache_name)

    def __get_lock(self, key):
        # pylint: disable=R1732
        self.__lock_map_lock.acquire()
        if key not in self.__lock_map:
            self.__lock_map[key] = threading.Lock()
        lock = self.__lock_map[key]
        self.__lock_map_lock.release()
        return lock

    def __acquire_lock(self, key):
        self.__get_lock(key).acquire()

    def __release_lock(self, key):
        self.__get_lock(key).release()

    def __get_cache_file_name(self, cache_key):
        return '%s/%s' % (
            self.__get_dir(),
            _Cache.__get_hash(cache_key),
        )

    def __get_file_exists(self, key):
        return os.path.exists(self.__get_cache_file_name(key))

    def __get_from_file(self, key):
        data_json = filex.read(self.__get_cache_file_name(key))

        if data_json == '':
            return None
        packet = json.loads(data_json)
        return packet

    def __set(self, key, data):
        self.__acquire_lock(key)
        packet = {
            'data': data,
            'set_time': time.time(),  # TODO: Change this later to use timex
        }
        filex.write(
            self.__get_cache_file_name(key),
            json.dumps(packet, ensure_ascii=True),
        )
        self.__store[key] = packet
        self.__release_lock(key)

    def get(self, key_or_list, fallback):
        """Get data from cache if cache key exists, if not from fallback.

        Args:
            key_or_list (str or list): the cache key can be given as a string,
                or a list of strings, which are joined
            fallback (function): function to call of cache key not in cache

        Returns:
            data

        """
        if isinstance(key_or_list, list):
            key = ':'.join(key_or_list)
        else:
            key = key_or_list

        packet = None

        if key in self.__store:
            packet = self.__store[key]

        if not packet:
            if self.__get_file_exists(key):
                packet = self.__get_from_file(key)
                if packet is not None:
                    self.__store[key] = packet

        if packet:
            if 'set_time' in packet:
                min_set_time = time.time() - self.__timeout
                if packet['set_time'] > min_set_time:
                    return packet['data']

        data = fallback()
        if data is not None:
            self.__set(key, data)
        return data

    def flush_store(self):
        """Flush all cache."""
        self.__store = {}

    def flush(self, key):
        """Flush cache for a given cache key.

        Args:
            key (str): Cache key to flush
        """
        self.__acquire_lock(key)
        if self.__get_file_exists(key):
            os.remove(self.__get_cache_file_name(key))
        if key in self.__store:
            del self.__store[key]
        self.__release_lock(key)


def cache(cache_name, timeout=CACHE_DEFAULT_TIMEOUT):
    """Wrap class Cache as decorator.

    Args:
        cache_name (str): cache name
        timeout (int): timeout in seconds

    .. code-block:: python

        from utils.cache import cache

        @cache('test', 86400)
        def long_operation():
            import time
            time.sleep(100)
            return 1

        >>> long_operation() # takes >100s to run
        >>> long_operation() # runs almost instantly

    """

    def cache_inner(func):

        @wraps(func)
        def cache_inner_inner(*args, **kwargs):
            cache_key = json.dumps({
                'cache_name': cache_name,
                'function_name': func.__name__,
                'kwargs': kwargs,
                'args': args,
            })

            def fallback():
                return func(*args, **kwargs)

            return _Cache(cache_name, timeout).get(cache_key, fallback)

        return cache_inner_inner
    return cache_inner
